/*!
 * \name        TLE5012B_SPI.h - Mbed port of Arduino library for the TLE5012B angle sensor.
 * \author      Infineon Technologies AG (Dr.Olaf Filies)
 * \copyright   2019 Infineon Technologies AG
 * \version     2.0.1
 * \brief       GMR-based angle sensor for angular position sensing in automotive applications
 * \details     Ported to Mbed by Zoltan Hudak 2020-08
 *
 * The TLE5012B is a 360° angle sensor that detects the orientation of a magnetic field.
 * This is achieved by measuring sine and cosine angle components with monolithic integrated
 * Giant Magneto Resistance (iGMR) elements. These raw signals (sine and cosine) are digitally
 * processed internally to calculate the angle orientation of the magnetic field (magnet).
 * The TLE5012B is a pre-calibrated sensor. The calibration parameters are stored in laser fuses.
 * At start-up the values of the fuses are written into flip-flops, where these values can be changed
 * by the application-specific parameters. Further precision of the angle measurement over a wide
 * temperature range and a long lifetime can be improved by enabling an optional internal autocalibration
 * algorithm. Data communications are accomplished with a bi-directional Synchronous Serial Communication (SSC)
 * that is SPI-compatible. The sensor configuration is stored in registers, which are accessible by the
 * SSC interface. Additionally four other interfaces are available with the TLE5012B: Pulse-Width-Modulation (PWM)
 * Protocol, Short-PWM-Code (SPC) Protocol, Hall Switch Mode (HSM) and Incremental Interface (IIF). These interfaces
 * can be used in parallel with SSC or alone. Pre-configured sensor derivates with different interface settings are available.
 * Online diagnostic functions are provided to ensure reliable operation.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
 * disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the copyright holders nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE  FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, _mosi, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY,OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "TLE5012B_SPI.h"

/**
 * @brief   Creates an SPI interface for SSC communication with TLE5012B sensor.
 * @note
 * @param
 * @retval
 */
TLE5012B_SPI::TLE5012B_SPI(PinName mosi, PinName miso, PinName sck) :
    _mosi(mosi),
    _sck(sck),
    _spi(mosi, miso, sck)
{ }

/**
 * @brief   Configures SPI interface for cummunication with TLE5012B chip.
 * @note
 * @param
 * @retval
 */
void TLE5012B_SPI::begin(DigitalOut* cs)
{
    *cs = 1;

    _spi.format(16, 1);         // 16 bits, MODE1, MSB first
    _spi.frequency(8000000);    // 8MHz (maximum supported by TLE5012B)
}

/**
 * @brief   Triggers an update.
 * @note
 * @param
 * @retval
 */
void TLE5012B_SPI::triggerUpdate()
{
    _sck = 0;
    _mosi = 1;
}

/**
 * @brief   Sends command(s) to the TLE5012B sensor and receives data.
 * @note
 * @param
 * @retval
 */
void TLE5012B_SPI::sendReceiveSpi(DigitalOut* cs, uint16_t* tx, uint16_t txLen, uint16_t* rx, uint16_t rxLen)
{
    // Send txData over SPI & receive dummy data
    *cs = 0;
    for (uint16_t i = 0; i < txLen; i++) {
        _spi.write(tx[i]);
    }

    wait_us(2);

    // Send dummy bytes over SPI & receive data
    for (uint16_t i = 0; i < rxLen; i++) {
        rx[i] = _spi.write(SPI_FILL_WORD);
    }
    *cs = 1;
}

/**
 * @brief
 * @note
 * @param
 * @retval
 */
inline void TLE5012B_SPI::writeBit(int value)
{
    _sck = 0;
    wait_us(1);
    _sck = 1;
    wait_us(1);
    if (value != 0) {
        _mosi = 1;
    }
    else {
        _mosi = 0;
    }

    wait_us(1);
    _sck = 0;
}

/**
 * @brief
 * @note
 * @param
 * @retval
 */
inline int TLE5012B_SPI::readBit()
{
    int out = 0;

    _sck = 0;
    wait_us(1);
    _sck = 1;

    wait_us(1);
    _sck = 0;
    if (_mosi.read() != 0) {
        out = 1;
    }

    return out;
}

/**
 * @brief
 * @note
 * @param
 * @retval
 */
void TLE5012B_SPI::write(uint16_t word)
{
    for (int i = 0; i < 16; ++i) {
        writeBit(word & 0x8000);
        word <<= 1;
    }
}

/**
 * @brief
 * @note
 * @param
 * @retval
 */
uint16_t TLE5012B_SPI::read()
{
    uint16_t    word = 0;

    for (int i = 0; i < 16; ++i) {
        word <<= 1;
        word |= readBit();
    }

    return word;
}
